import datetime
import json
from dataclasses import dataclass, field
from colorama import Fore, Style, Back
import re
import numpy as np
import h5py

#TODO: Save only certain log levels
#TODO: Autosave
#TODO: Log more info
#TODO: Log to string etc
#TODO: Integrate with logger

xDEBUG = -10
xINFO = -20
xWARNING = -30
xERROR = -40
xCRITICAL = -50

RECORD = -25
CORE = -30

NOTSET = 0
DEBUG = 10		# Used for debugging
INFO = 20		# Used for reporting basic high-level program functioning (that does not involve an error)
WARNING = 30 	# Warning for software
ERROR = 40		# Software error
CRITICAL = 50	# Critical error

class SortConditions:
	
	time_start = None
	time_end = None
	contains_and = []
	contains_or = []
	index_start = None
	index_end = None
	

#TODO: Make the keys in color_overrides match the variables in LogEntry (currently undefined)
@dataclass
class LogFormat:
	
	show_detail:bool = False
	use_color:bool = True
	default_color:dict = field(default_factory=lambda: {"main": Fore.WHITE+Back.RESET, "bold": Fore.LIGHTBLUE_EX, "quiet": Fore.LIGHTBLACK_EX, "alt": Fore.YELLOW, "label": Fore.GREEN})
	color_overrides:dict = field(default_factory=lambda: {DEBUG: {"label": Fore.LIGHTBLACK_EX},
														INFO: {},
														WARNING: {"label": Fore.YELLOW},
														ERROR: {"label": Fore.LIGHTRED_EX},
														CRITICAL: {"label": Fore.RED},
														RECORD: {"label": Fore.CYAN},
														CORE: {"label": Fore.CYAN}
	})
	detail_indent:str = "\t "
	strip_newlines:bool = True

def str_to_level(lvl:str) -> int:
	
	# Set level
	if lvl == "DEBUG":
		return DEBUG
	elif lvl == "RECORD":
		return RECORD
	elif lvl == "INFO":
		return INFO
	elif lvl == "CORE":
		return CORE
	elif lvl == "WARNING":
		return WARNING
	elif lvl == "ERROR":
		return ERROR
	elif lvl == "CRITICAL":
		return CRITICAL
	else:
		return False

class LogEntry:
	
	default_format = LogFormat()
	
	def __init__(self, level:int=0, message:str="", detail:str=""):
		
		# Set timestamp
		self.timestamp = datetime.datetime.now()
		
		if detail is None:
			detail = ""
		if message is None:
			message = ""
		
		# Set level
		if level not in [DEBUG, INFO, WARNING, ERROR, CRITICAL]:
			self.level = INFO
		else:
			self.level = level
		
		# Set message
		self.message = message
		self.detail = detail
	

	
	def init_dict(self, data_dict:dict) -> bool:
		
		# Extract data from dict
		try:
			lvl = data_dict['level']
			msg = data_dict['message']
			dtl = data_dict['detail']
			ts = data_dict['timestamp']
		except:
			return False
		
		# Set level
		self.level = str_to_level(lvl)
		if self.level is None:
			return False
		
		self.message = msg # Set message
		self.detail = dtl
		self.timestamp = datetime.datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f')
		
		return True
	
	def get_level_str(self):
		
		if self.level == DEBUG:
			return "DEBUG"
		elif self.level == RECORD:
			return "RECORD"
		elif self.level == INFO:
			return "INFO"
		elif self.level == CORE:
			return "CORE"
		elif self.level == WARNING:
			return "WARNING"
		elif self.level == ERROR:
			return "ERROR"
		elif self.level == CRITICAL:
			return "CRITICAL"
		else:
			return "??"
		
	def get_dict(self):
		return {"message":self.message, "detail":self.detail, "timestamp":str(self.timestamp.strftime('%Y-%m-%d %H:%M:%S.%f')) , "level":self.get_level_str()}
	
	def get_json(self):
		return json.dumps(self.get_dict())
	
	def str(self, str_fmt:LogFormat=None) -> str:
		''' Represent the log entry as a string.'''
		
		# Get format specifier
		if str_fmt is None:
			str_fmt = LogEntry.default_format
		
		# Apply or wipe colors
		if str_fmt.use_color:
			c_main = str_fmt.default_color['main']
			c_bold = str_fmt.default_color['bold']
			c_quiet = str_fmt.default_color['quiet']
			c_alt = str_fmt.default_color['alt']
			c_label = str_fmt.default_color['label']
			
			# Apply log-level color-overrides
			if self.level in str_fmt.color_overrides:
				if 'main' in str_fmt.color_overrides[self.level]:
					c_main = str_fmt.color_overrides[self.level]['main']
				if 'bold' in str_fmt.color_overrides[self.level]:
					c_bold = str_fmt.color_overrides[self.level]['bold']
				if 'quiet' in str_fmt.color_overrides[self.level]:
					c_quiet = str_fmt.color_overrides[self.level]['quiet']
				if 'alt' in str_fmt.color_overrides[self.level]:
					c_alt = str_fmt.color_overrides[self.level]['alt']
				if 'label' in str_fmt.color_overrides[self.level]:
					c_label = str_fmt.color_overrides[self.level]['label']
				
		else:
			c_main = ''
			c_bold = ''
			c_quiet = ''
			c_alt = ''
			c_label = ''
		
		# If requested, remove all newlines
		if str_fmt.strip_newlines:
			message = self.message.replace("\n", "")
			detail = self.detail.replace("\n", "")
			
		
		# Create base string
		s = f"{c_alt}[{c_label}{self.get_level_str()}{c_alt}]{c_main} {markdown(message, str_fmt)} {c_quiet}| {self.timestamp}{Style.RESET_ALL}"
		
		# Add detail if requested
		if str_fmt.show_detail and len(detail) > 0:
			s = s + f"\n{str_fmt.detail_indent}{c_quiet}{detail}"
		
		return s
	
	def matches_sort(self, orders:SortConditions) -> bool:
		''' Checks if the entry matches the conditions specified by the SortConditions 'orders'. Returns true if they match and false if they don't. NOTE: Does not check index, that is only valid in a LogPile context.
		'''
		# Check if time conditions are specified
		if (orders.time_start is not None) and (orders.time_end is not None):
			
			# Check if conditions agree
			if self.timestamp < orders.time_start or self.timestamp > orders.time_end:
				return False
		
		# Check if contains_and is specified
		if len(orders.contains_and) > 0:
			
			# Check if all are hits
			for targ in orders.contains_and:
				if (targ not in self.message) and (targ not in self.detail):
					return False
		
		# Check if contains_or is specified
		if len(orders.contains_or) > 0:
			
			found_or = False
			
			# Check if any are hits
			for targ in orders.contains_or:
				if (targ in self.message) or (targ in self.detail):
					found_or = True
					break
			
			# Return negative if none matched
			if not found_or:
				return False
		
		# All matched!
		return True
	
def markdown(msg:str, str_fmt:LogFormat=None) -> str:
	""" Applys Pylogfile markdown
		> Temporarily change to bold
		< Revert to previous color
		
		>:n Temporariliy change to color 'n'. n-codes: Case insensitive
			1 or m: Main
			2 or b: Bold
			3 or q: Quiet
			4 or a: Alt
			5 or l: Label
		
		>> Permanently change to bold
		>>:n Permanently change to color n
		
		\\>, \\<, Type character without color adjustment. So to get >>:3
			to appear you'd type \\>\\>:3.
		
		If you want to type > followed by a character
		
	"""
	
	# Get default format
	if str_fmt is None:
		str_fmt = LogEntry.default_format
	
	# Apply or wipe colors
	if str_fmt.use_color:
		c_main = str_fmt.default_color['main']
		c_bold = str_fmt.default_color['bold']
		c_quiet = str_fmt.default_color['quiet']
		c_alt = str_fmt.default_color['alt']
		c_label = str_fmt.default_color['label']
	else:
		c_main = ''
		c_bold = ''
		c_quiet = ''
		c_alt = ''
		c_label = ''
	
	# This is the color that a return character will restore
	return_color = c_main
	
	# Get every index of '>', '<', and '\\'
	idx = 0
	replacements = []
	while idx < len(msg):
		
		# Look for escape character
		if msg[idx] == '\\':
			
			# If next character is > or <, remove the escape
			if idx+1 < len(msg) and msg[idx+1] == '>':
				replacements.append({'text': '>', 'idx_start': idx, 'idx_end': idx+1})
			elif idx+1 < len(msg) and msg[idx+1] == '<':
				replacements.append({'text': '<', 'idx_start': idx, 'idx_end': idx+1})
			
			idx += 2 # Skip next character - restart
			continue
		
		# Look for non-escaped >
		elif msg[idx] == '>':
			
			idx_start = idx
			is_permanent = False
			color_spec = c_bold
			is_invalid = False
			
			# Check for permanent change
			if idx+1 < len(msg) and msg[idx+1] == '>': # Permanent change
				is_permanent = True
				idx += 1
			
			# Check for color specifier
			if idx+2 < len(msg) and msg[idx+1] == ':': # Found color specifier
				
				if msg[idx+2].upper() in ['1', 'M']:
					color_spec = c_main
				elif msg[idx+2].upper() in ['2', 'B']:
					color_spec = c_bold
				elif msg[idx+2].upper() in ['3', 'Q']:
					color_spec = c_quiet
				elif msg[idx+2].upper() in ['4', 'A']:
					color_spec = c_alt
				elif msg[idx+2].upper() in ['5', 'L']:
					color_spec = c_label
				else:
					# Unrecognized code, do not modify
					is_invalid = True
				
				idx += 2
			
			# Apply changes and text replacements
			if not is_invalid:
				replacements.append({'text': color_spec, 'idx_start': idx_start, 'idx_end':idx})
				
				# If permanent apply change
				if is_permanent:
					return_color = color_spec
		
		# Look for non-escaped <
		elif msg[idx] == '<':
			
			replacements.append({'text': return_color, 'idx_start': idx, 'idx_end': idx})
		
		# Increment counter
		idx += 1
		
	# Apply replacements
	rich_msg = msg
	for rpl in reversed(replacements):
		rich_msg = rich_msg[:rpl['idx_start']] + rpl['text'] + rich_msg[rpl['idx_end']+1:]
	
	return rich_msg
		

class LogPile:
	
	JSON = "format-json"
	TXT = "format-txt"
	
	def __init__(self, filename:str="", autosave:bool=False, str_fmt:LogFormat=None):
		
		# Initialize format with defautl
		if str_fmt is None:
			str_fmt = LogFormat()
		
		self.terminal_output_enable = True
		self.terminal_output_details = False
		self.terminal_level = INFO
		
		self.autosave_enable = autosave
		self.filename = filename
		self.autosave_period_s = 300
		self.autosave_level = INFO
		self.autosave_format = LogPile.JSON
		
		self.str_format = str_fmt
		
		self.logs = []
	
	def debug(self, message:str, detail:str=""):
		''' Logs data at DEBUG level. '''
		
		self.add_log(DEBUG, message, detail=detail)
	
	def info(self, message:str, detail:str=""):
		''' Logs data at INFO level. '''
		
		self.add_log(INFO, message, detail=detail)
	
	def warning(self, message:str, detail:str=""):
		''' Logs data at WARNING level. '''
		
		self.add_log(WARNING, message, detail=detail)
	
	def error(self, message:str, detail:str=""):
		''' Logs data at ERROR level. '''
		
		self.add_log(ERROR, message, detail=detail)

	def critical(self, message:str, detail:str=""):
		''' Logs data at CRITICAL level. '''
		
		self.add_log(CRITICAL, message, detail=detail)
	
	def add_log(self, level:int, message:str, detail:str=""):
		
		# Create new log object
		nl = LogEntry(level, message, detail=detail)
		
		# Add to list
		self.logs.append(nl)
		
		# Process new log with any auto-running features
		self.run_new_log(nl)
	
	def run_new_log(self, nl:LogEntry):
		
		# Print to terminal
		if self.terminal_output_enable:
			print(f"{nl.str(self.str_format)}{Style.RESET_ALL}")
	
	def to_dict(self):
		return [x.get_dict() for x in self.logs]
	
	def save_json(self, save_filename:str):
		''' Saves all log data to a JSON file '''
		
		ad = self.to_dict()
		
		# Open file
		with open(save_filename, 'w') as fh:
			json.dump({"logs":ad}, fh, indent=4)
	
	def load_json(self, read_filename:str):
		
		all_success = True
		
		# Read JSON dictionary
		with open(read_filename, 'r') as fh:
			ad = json.load(fh)
		
		# Populate logs
		for led in ad['logs']:
			nl = LogEntry()
			if nl.init_dict(led):
				self.logs.append(nl)
			else:
				all_success = False
		
		return all_success
	
	def save_hdf(self, save_filename):
		
		ad = self.to_dict()
		
		message_list = []
		detail_list = []
		timestamp_list = []
		level_list = []
		
		# Create HDF data types
		for de in ad:
			
			message_list.append(de['message'])
			detail_list.append(de['detail'])
			timestamp_list.append(de['timestamp'])
			level_list.append(de['level'])
		
		# Write file
		with h5py.File(save_filename, 'w') as fh:
			fh.create_group("logs")
			fh['logs'].create_dataset('message', data=message_list)
			fh['logs'].create_dataset('detail', data=detail_list)
			fh['logs'].create_dataset('timestamp', data=timestamp_list)
			fh['logs'].create_dataset('level', data=level_list)
	
	def load_hdf(self, read_filename:str):
		
		all_success = True
		
		# Load file contents
		with h5py.File(read_filename, 'r') as fh:
			message_list = fh['logs']['message'][()]
			detail_list = fh['logs']['detail'][()]
			timestamp_list = fh['logs']['timestamp'][()]
			level_list = fh['logs']['level'][()]
		
		# Convert to dictionary
		for nm,nd,nt,nl in zip(message_list, detail_list, timestamp_list, level_list):
			
			# Create dictionary
			dd = {'message': nm.decode('utf-8'), 'detail':nd.decode('utf-8'), 'timestamp': nt.decode('utf-8'), 'level':nl.decode('utf-8')}
			
			# Create LogEntry
			nl = LogEntry(message=nm, detail=nd)
			if nl.init_dict(dd):
				self.logs.append(nl)
			else:
				all_success = False
		
		return all_success
			
	
	def save_txt(self):
		pass
	
	def begin_autosave(self):
		pass
	
	def show_logs(self, min_level:int=DEBUG, max_level:int=CRITICAL, max_number:int=None, from_beginning:bool=False, show_index:bool=True, sort_orders:SortConditions=None, str_fmt:LogFormat=None):
		'''
		Shows logs matching the specified conditions
		
		Args:
			min_level (int): Minimum logging level to display
			max_level (int): Maximum logging level to display
			max_number (int): Maximum number of logs to show
			from_beginning (bool): Show logs starting from beginning.
			show_index (bool): Show or hide the index of the log entry by each entry.
		
		Returns:
			None
		'''
		
		# Check max number is zero or less
		if max_number is not None and max_number < 1:
			return
		
		# Get list order
		if not from_beginning:
			log_list = reversed(self.logs)
			idx_list = reversed(list(np.linspace(0, len(self.logs)-1, len(self.logs))))
		else:
			log_list = self.logs
			idx_list = list(np.linspace(0, len(self.logs)-1, len(self.logs)))
		
		# Scan over logs
		idx_str = ""
		for idx, lg in zip(idx_list, log_list):
			
			# Check log level
			if lg.level < min_level or lg.level > max_level:
				continue
			
			# If sort orders are provided, perform check
			if (sort_orders is not None):
				
				# If time and contents searches dont hit, skip entry
				if (not lg.matches_sort(sort_orders)):
					continue
				
				# Check if index filter is requested
				if (sort_orders.index_start is not None) and (sort_orders.index_end is not None):
					
					# If entry doesn't hit, skip it
					if (idx < sort_orders.index_start) or (idx > sort_orders.index_end):
						continue
			
			# Print log
			if show_index:
				# idx_str = f"{Fore.LIGHTBLACK_EX}[{Fore.YELLOW}{int(idx)}{Fore.LIGHTBLACK_EX}] "
				idx_str = f"{Fore.WHITE}[{Fore.WHITE}{int(idx)}{Fore.WHITE}] "
			
			if str_fmt is None:
				print(f"{idx_str}{lg.str(self.str_format)}{Style.RESET_ALL}")
			else:
				print(f"{idx_str}{lg.str(str_fmt)}{Style.RESET_ALL}")
			
			# Run counter if specified
			if max_number is not None:
				
				# Decrement
				max_number -= 1
				
				# Check for end
				if max_number < 1:
					cq = self.str_format.default_color['quiet']
					print(f"\t{cq}...{Style.RESET_ALL}")
					break


#!/usr/bin/env python

import os
import sys
from colorama import Fore, Style
from pylogfile import *
import argparse
from itertools import groupby, count, filterfalse
import dataclasses

try:
	
	# Requires Python >= 3.9
	import importlib.resources
	
	mod_path = importlib.resources.files(__package__)
	inp_file = (mod_path / 'assets' / 'lumberjack_help.json')
	with inp_file.open("r") as f:  # or "rt" as text file with universal newlines
		file_contents = f.read()
	
	help_data = json.loads(file_contents)
except AttributeError:
	help_data = {}
	print(f"{Fore.LIGHTRED_EX}Upgrade to Python >= 3.9 for access to importlib and CLI help data.")


def barstr(text:str, width:int=80, bc:str='*', pad:bool=True):

		s = text

		# Pad input if requested
		if pad:
			s = " " + s + " "

		pad_back = False
		while len(s) < width:
			if pad_back:
				s = s + bc
			else:
				s = bc + s
			pad_back = not pad_back

		return s

#TODO: Change where elipses print for --last

#TODO: Make it so you can apply search strings to only details or only message or both
#TODO: Modify keyword search to preserve strings so phrass can be searched
#TODO: Search by timestamp
#TODO: Search by index
#TODO: Automatically sort help --list keys (port to * as well)
#TODO: Options to turn On/Off: Elipses between incongruous logs, log index #s, color-print, option to pad index number so always same number of digits

##================================================================
# Read commandline Arguments

parser = argparse.ArgumentParser()
parser.add_argument('filename')
parser.add_argument('--last', help="Show last X number of logs", action='store_true')
parser.add_argument('--first', help="Show first X number of logs", action='store_true')
parser.add_argument('-g', '--gui', help="Use graphical interface", action='store_true')
parser.add_argument('-a', '--all', help="Print all logs", action='store_true')
parser.add_argument('-nc', '--nocli', help="Skip the CLI", action='store_true')
args = parser.parse_args()

if args.gui:
	print(f"{Fore.RED}GUI has not been implemented. Continuing with CLI.{Style.RESET_ALL}")

##================================================================


class LJSettings:
	
	def __init__(self, s=None):
		self.from_beginning:bool = True
		self.num_print:int = 15
		self.min_level = DEBUG
		self.max_level = CRITICAL
		self.show_detail = False
		
		if s is not None:
			self.get_vals(s)
	
	def get_vals(self, s):
		self.from_beginning = s.from_beginning
		self.num_print = s.num_print
		self.min_level = s.min_level
		self.max_level = s.max_level
	
def str_to_bool(val:str, strict:bool=True):
	''' Converts the string 0/1 or ON/OFF or TRUE/FALSE to a boolean '''
	
	if ('1' in val) or ('ON' in val.upper()) or ('TRUE' in val.upper()):
		return True
	elif ('0' in val) or ('OFF' in val.upper()) or ('FALSE' in val.upper()):
		return False
	else:
		if strict:
			return None
		else:
			return False

class StringIdx():
	def __init__(self, val:str, idx:int, idx_end:int=-1):
		self.str = val
		self.idx = idx
		self.idx_end = idx_end;

	def __str__(self):
		return f"[{self.idx}]\"{self.str}\""

	def __repr__(self):
		return self.__str__()

# Looks for characters in 'delims' in string 'input'. Supposing the string is to
# be broken up at each character in 'delims', the function returns a generator
# with the start and end+1 indecies of each section.
#
def parseTwoIdx(input:str, delims:str):
	p = 0
	for k, g in groupby(input, lambda x:x in delims):
		q = p + sum(1 for i in g)
		if not k:
			yield (p, q) # or p, q-1 if you are really sure you want that
		p = q

def parseIdx(input:str, delims:str=" ", keep_delims:str=""):
	""" Parses a string, breaking it up into an array of words. Separates at delims. """
	
	out = []
	
	sections = list(parseTwoIdx(input, delims))
	for s in sections:
		out.append(StringIdx(input[s[0]:s[1]], s[0], s[1]))
	return out

def ensureWhitespace(s:str, targets:str, whitespace_list:str=" \t", pad_char=" "):
	""" """
	
	# Remove duplicate targets
	targets = "".join(set(targets))
	
	# Add whitespace around each target
	for tc in targets:
		
		start_index = 0
		
		# Find all instances of target
		while True:
			
			# Find next instance of target
			try:
				idx = s[start_index:].index(tc)
				idx += start_index
			except ValueError as e:
				break # Break when no more instances
			
			# Update start index
			start_index = idx + 1
			
			# Check if need to pad before target
			add0 = True
			if idx == 0:
				add0 = False
			elif s[idx-1] in whitespace_list:
				add0 = False
			
			# Check if need to pad after target
			addf = True
			if idx >= len(s)-1:
				addf = False
			elif s[idx+1] in whitespace_list:
				addf = False
			
			# Add required pad characters
			if addf:
				s = s[:idx+1] + pad_char + s[idx+1:]
				start_index += 1 # Don't scan pad characters
			if add0:
				s = s[:idx] + pad_char + s[idx:]
	return s

def show_help():
	print(f"{Fore.RED}Requires name of file to analyze.{Style.RESET_ALL}")

def main():
	
	settings = LJSettings()
	
	# Create logpile object
	log = LogPile()
	
	# Create format object
	fmt = LogFormat()
	fmt.show_detail = True
	
	# Get filename from arguments
	filename = args.filename
	
	# Read file
	if filename[-4:].upper() == ".HDF":
		if not log.load_hdf(filename):
			print("\tFailed to read HDF")
	elif filename[-5:].upper() == ".JSON":
		if not log.load_hdf(filename):
			print("\tFailed to read JSON file.")
	
	# Show all logs if requested
	if args.all:
		log.show_logs()
	elif args.first:
		log.show_logs(max_number=settings.num_print, from_beginning=True, min_level=settings.min_level, max_level=settings.max_level)
	elif args.last:
		log.show_logs(max_number=settings.num_print, min_level=settings.min_level, max_level=settings.max_level)
	
	# Run CLI
	running = not args.nocli
	while running:
		
		# Get user input
		cmd_raw = input(f"{Fore.GREEN}LUMBERJACK > {Style.RESET_ALL}")
		
		# Break into words
		words = parseIdx(cmd_raw, " \t")
		
		# Skip empty lines
		if len(words) < 1:
			continue
		
		# Get command string
		cmd = words[0].str.upper()
		
		# Process commands
		if cmd == "EXIT":
			running = False
		elif cmd == "CLS" or cmd == "CLEAR":
			if os.name == 'nt':
				os.system("cls")
			else:
				os.system("clear")
		elif cmd == "MIN-LEVEL":
			
			# Check number of arguments
			if len(words) < 2:
				print(f"{Fore.LIGHTRED_EX}MIN-LEVEL requires a level to be specified.{Style.RESET_ALL}")
				continue
			
			# Get level int
			lvl_str = words[1].str.upper()
			lvl_int = str_to_level(lvl_str)
			if lvl_int is None:
				print(f"{Fore.LIGHTRED_EX}Unrecognized level spcifier '{lvl_str}'.{Style.RESET_ALL}")
				continue
			
			# Assign min level
			min_level = lvl_int
		elif cmd == "MAX-LEVEL":
			
			# Check number of arguments
			if len(words) < 2:
				print(f"{Fore.LIGHTRED_EX}MAX-LEVEL requires a level to be specified.{Style.RESET_ALL}")
				continue
			
			# Get level int
			lvl_str = words[1].upper()
			lvl_int = str_to_level(lvl_str)
			if lvl_int is None:
				print(f"{Fore.LIGHTRED_EX}Unrecognized level spcifier '{lvl_str}'.{Style.RESET_ALL}")
				continue
			
			# Assign min level
			max_level = lvl_int
		elif cmd == "SHOW":
			
			# Initialize local copies of show parameters, for local overrides
			local_settings = LJSettings(settings)
			local_fmt = dataclasses.replace(log.str_format)
			show_all = False
			search_orders = SortConditions()
			do_search = False
			
			# Check for flags
			idx = 1
			while idx < len(words):
				if words[idx].str == "-n" or words[idx].str == "--num":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--num' requires an argument (the number of logs to display).")
						break
					
					# Try to interpret argument
					try:
						local_settings.num_print = int(words[idx+1].str)
					except:
						w2 = words[idx+1]
						print(f"{Fore.LIGHTRED_EX}Failed to interpret number provided, '{w2.str}'.{Style.RESET_ALL}")
						idx += 1
						continue
					idx += 1
					
					# Check if num_print is zero (inf), and reassign to None
					if local_settings.num_print == 0:
						local_settings.num_print = None
					
				elif words[idx].str == "-a" or words[idx].str == "--all":
					show_all = True
				elif words[idx].str == "-f" or words[idx].str == "--first":
					local_settings.from_beginning = True
				elif words[idx].str == "-l" or words[idx].str == "--last":
					local_settings.from_beginning = False
				elif words[idx].str == "-d" or words[idx].str == "--detail":
					
					# Verify argument is present
					if idx+1 >= len(words):
						
						# If no argument is present, default to turning detail ON
						local_fmt.show_detail = True
					
					# Get level int
					d_val = str_to_bool(words[idx+1].str)
					if d_val is None:
						s = words[idx+1].str
						print(f"{Fore.LIGHTRED_EX}Unrecognized boolean spcifier '{s}'.{Style.RESET_ALL}")
						idx += 1
						continue
					idx += 1
					
					# Assign value
					local_fmt.show_detail = d_val
				elif words[idx].str == "-i" or words[idx].str == "--index":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--index' requires an argument (The index or index range 'START:END' to display).")
						break
					
					# Interpret argument
					idx_colon = words[idx+1].str.find(':') # Look for colon
					if idx_colon == -1: # No colon present
						try:
							search_orders.index_start = int(words[idx+1].str)
							search_orders.index_end = search_orders.index_start
						except Exception as e:
							s = words[idx+1].str
							print(f"{Fore.LIGHTRED_EX}Unrecognized index spcifier '{s}'. ({e}){Style.RESET_ALL}")
							idx += 1
							search_orders.index_start = None
							search_orders.index_end = None
							continue
					else: # Colon was present - interpret range
						try:
							search_orders.index_start = int(words[idx+1].str[:idx_colon])
							search_orders.index_end = int(words[idx+1].str[idx_colon+1:])
							local_settings.num_print = None
						except Exception as e:
							s1 = words[idx+1].str[:idx_colon]
							s2 = words[idx+1].str[idx_colon+1:]
							print(f"{Fore.LIGHTRED_EX}Failed to interpret index range '{s1}' to '{s2}'. ({e}){Style.RESET_ALL}")
							idx += 1
							search_orders.index_start = None
							search_orders.index_end = None
							continue
					idx += 1
					do_search = True
					
				elif words[idx].str == "-m" or words[idx].str == "--min":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--min' requires an argument (the minimum log level to display).")
						break
					
					# Get level int
					lvl_str = words[idx+1].str.upper()
					lvl_int = str_to_level(lvl_str)
					if lvl_int is None:
						print(f"{Fore.LIGHTRED_EX}Unrecognized level spcifier '{lvl_str}'.{Style.RESET_ALL}")
						idx += 1
						continue
					idx += 1
					
					# Assign value
					local_settings.min_level = lvl_int
					
				elif words[idx].str == "-x" or words[idx].str == "--max":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--max' requires an argument (the maximum log level to display).")
						break
					
					# Get level int
					lvl_str = words[idx+1].str.upper()
					lvl_int = str_to_level(lvl_str)
					if lvl_int is None:
						print(f"{Fore.LIGHTRED_EX}Unrecognized level spcifier '{lvl_str}'.{Style.RESET_ALL}")
						idx += 1
						continue
					idx += 1
					
					# Assign value
					local_settings.max_level = lvl_int
					
				elif words[idx].str == "-c" or words[idx].str == "--contains":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--contains' requires an argument (the string to search for).")
						break
					
					# Get level int
					search_orders.contains_or.append(words[idx+1].str)
					idx += 1
					do_search = True
				
				elif words[idx].str == "-cc" or words[idx].str == "--andcontains":
					
					# Verify argument is present
					if idx+1 >= len(words):
						print(f"{Fore.LIGHTRED_EX}Flag '--contains' requires an argument (the string to search for).")
						break
					
					# Get level int
					search_orders.contains_and.append(words[idx+1].str)
					idx += 1
					do_search = True
				
				idx += 1
			
			# If no searches were previded, set object to None
			if not do_search:
				search_orders = None
			
			# Show logs using local settings
			if show_all:
				log.show_logs(from_beginning=True, str_fmt=local_fmt)
			else:
				log.show_logs(max_number=local_settings.num_print, from_beginning=local_settings.from_beginning, min_level=local_settings.min_level, max_level=local_settings.max_level, sort_orders=search_orders, str_fmt=local_fmt)
				
		elif cmd == "NUM-PRINT":
			
			# Check number of arguments
			if len(words) < 2:
				print(f"{Fore.LIGHTRED_EX}NUM-PRINT requires a second argument (number of logs to print).{Style.RESET_ALL}")
				continue
			
			# interpret argument
			try:
				head_len = int(words[1].str)
			except Exception as e:
				w2 = words[1]
				print(f"{Fore.LIGHTRED_EX}Failed to interpret number provided, '{w2.str}' ({e}).{Style.RESET_ALL}")
		elif cmd == "HELP":
			
			HELP_WIDTH = 80
			TABC = "    "
			
			color1 = Fore.WHITE # Body text
			color2 = Fore.LIGHTYELLOW_EX # Titles/headers
			color3 = Fore.YELLOW # Type specifiers, etc
			color4 = Fore.LIGHTBLACK_EX # brackets and accents
			
			hstr = ""
			
			list_commands = False
			
			# Check for flags
			print_long = False
			if len(words) > 1:
				for tk in words:
					if tk.str == "-l" or tk.str == "--list":
						list_commands = True

			
			if list_commands:
				
				# title
				hstr += color2 + "-"*HELP_WIDTH + Style.RESET_ALL + "\n"
				hstr += color2 + barstr(f"ALL COMMANDS", HELP_WIDTH, "-", pad=True) + Style.RESET_ALL + "\n\n"
				
				for cmd in help_data.keys():
					desc = help_data[cmd]['description']
					hstr += f"{TABC}{Fore.CYAN}{cmd}{color1}: {desc}\n"
				
				print(hstr)
				continue
			
			# Check for number of arguments
			if len(words) < 2:
				hcmd = "HELP"
			else:
				hcmd = words[1].str.upper()
			
			cmd_list = help_data.keys()
			
			if hcmd in cmd_list: # HCMD is a COMMAND name
			
				## Print help data:
				try:
					# title
					hstr += color2 + "-"*HELP_WIDTH + Style.RESET_ALL + "\n"
					hstr += color2 + barstr(f"{hcmd} Help", HELP_WIDTH, "-", pad=True) + Style.RESET_ALL + "\n\n"
					
					# Description
					hstr += f"{color2}Description:\n"
					hstr += f"{color1}{TABC}" + help_data[hcmd]['description']+Style.RESET_ALL + "\n"
					
					# Arguments
					if len(help_data[hcmd]['arguments']) > 0:
						hstr += f"{color2}\nArguments:\n"
						for ar in help_data[hcmd]['arguments']:
							
							arg_name = ar['name']
							if ar['type'] in ["num", "str", "list", "cell", "list[cell]"]:
								type_name = ar['type']
							else:
								type_name = f"{Fore.RED}???"
							
							if ar['optional']:
								hstr += TABC + f"{color4}( {color1}{arg_name} {color4}[{color3}{type_name}{color4}]) "
							else:
								hstr += TABC + f"{color4}< {color1}{arg_name} {color4}[{color3}{type_name}{color4}]> "
							
							hstr += color1 + ar['description'] + "\n"
					
					# Flags
					if len(help_data[hcmd]['flags']) > 0:
						hstr += f"{color2}\nFlags:\n"
						for ar in help_data[hcmd]['flags']:
							
							if ar["short"] != "" and ar["long"] != "":
								hstr += TABC + color1 + ar['short'] + f"{color4}," + color1 + ar["long"] + color4 + ": "
							elif ar['short'] != "":
								hstr += TABC + color1 + ar['short'] + color4 + ": "
							else:
								hstr += TABC + color1 + ar['long'] + color4 + ": "
								
							
							hstr += color1 + ar['description'] + "\n"
					
					# Examples
					if len(help_data[hcmd]['examples']) > 0:
						hstr += f"{color2}\nExamples:\n"
						for ex_no, ar in enumerate(help_data[hcmd]['examples']):
							
							hstr += f"{color1}{TABC}Ex {ex_no}:\n"
							hstr += TABC + TABC + color4 + ">> " + color3 + ar['command'] + "\n"
							hstr += TABC + TABC + color1 + "Desc: " + color1 + ar['description'] + "\n"
						
						hstr += "\n"
					
					
					# See also
					if len(help_data[hcmd]['see_also']) > 0:
						hstr += f"{color2}\nSee Also:\n{TABC}{color1}"
						add_comma = False
						for ar in help_data[hcmd]['see_also']:
							
							if ar.upper() in cmd_list:
								
								if add_comma:
									hstr += ", "
								
								hstr += ar
								add_comma = True
					
					print(hstr)
				except Exception as e:
					print(f"Corrupt help data for selected entry '{hcmd}' ({e}).")
			
		elif cmd == "STATE":
			print(f"{Fore.CYAN}Lumberjack-CLI State:{Style.RESET_ALL}")
			print(f"    {Fore.YELLOW}Min. level: {Style.RESET_ALL}{settings.min_level}")
			print(f"    {Fore.YELLOW}Max. level: {Style.RESET_ALL}{settings.max_level}")
			print(f"    {Fore.YELLOW}Num. print: {Style.RESET_ALL}{settings.num_print}")
			print(f"    {Fore.YELLOW}Print from beginning: {Style.RESET_ALL}{settings.from_beginning}")
		elif cmd == "INFO":
			
			long_mode = False 
			
			# Check for additional arguments
			if len(words) > 1:
				
				# Scan over words
				for c in words[1:]:
					if c.str.upper() == "-L" or c.str.upper() == "--LONG":
						long_mode = True
					else:
						print(f"Unrecognized option '{c.str.upper()}'. Ignoring it.")
			
			print(f"{Fore.CYAN}Log Info: {Fore.LIGHTBLACK_EX}{filename}{Style.RESET_ALL}")
			print(f"    {Fore.YELLOW}number of Logs: {Style.RESET_ALL}{len(log.logs)}")
			if long_mode:
				# Count logs at each level
				ndebug, ninfo, nwarning, nerror, ncritical, nother = 0, 0, 0, 0, 0, 0
				for l in log.logs:
					if l.level == DEBUG:
						ndebug += 1
					elif l.level == INFO:
						ninfo += 1
					elif l.level == WARNING:
						nwarning += 1
					elif l.level == ERROR:
						nerror += 1
					elif l.level == CRITICAL:
						ncritical += 1
					else:
						nother += 1
				print(f"        {Fore.LIGHTBLACK_EX}Number of DEBUG: {Style.RESET_ALL}{ndebug}")
				print(f"        {Fore.LIGHTBLACK_EX}Number of INFO: {Style.RESET_ALL}{ninfo}")
				print(f"        {Fore.LIGHTBLACK_EX}Number of WARNING: {Style.RESET_ALL}{nwarning}")
				print(f"        {Fore.LIGHTBLACK_EX}Number of ERROR: {Style.RESET_ALL}{nerror}")
				print(f"        {Fore.LIGHTBLACK_EX}Number of CRITICAL: {Style.RESET_ALL}{ncritical}")
				print(f"        {Fore.LIGHTBLACK_EX}Number of other: {Style.RESET_ALL}{nother}")
			t_elapsed = log.logs[-1].timestamp - log.logs[0].timestamp
			print(f"    {Fore.YELLOW}Timespan: {Style.RESET_ALL}{t_elapsed}")
			if long_mode:
				ts_0 = log.logs[0].timestamp
				ts_1 = log.logs[-1].timestamp
				print(f"        {Fore.LIGHTBLACK_EX}First timestamp: {Style.RESET_ALL}{ts_0}")
				print(f"        {Fore.LIGHTBLACK_EX}Last timestamp: {Style.RESET_ALL}{ts_1}")
			# print(f"    {Fore.YELLOW}MAX-LEVEL: {Style.RESET_ALL}{max_level}")
			# print(f"    {Fore.YELLOW}NUM-PRINT: {Style.RESET_ALL}{head_len}")
		else:
			print(f"{Fore.LIGHTRED_EX}Unrecognized command '{cmd}'.{Style.RESET_ALL}")
		